from .bollyflix import BollyFlixME
from .fourkhdhub import FourkhdhubME
from .hdhub import Hdhub4uME
from .moviesmod import MoviesModME
from .vegamovies import VegaMoviesME

ALL = [BollyFlixME, Hdhub4uME, VegaMoviesME, FourkhdhubME, MoviesModME]
